----------------------------------------------------------------------------
-- |
-- Module      :  CodeOpt
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Perform simple code optimizations
--
-----------------------------------------------------------------------------

module CodeOpt where

import Type
import Core(Const(..))
import qualified Core as C
import qualified Data.Map as M
import Data.Maybe
import Data.Char
import Data.List
import Debug.Trace 
import Control.Monad
import Control.Monad.State
import Util
import Code
import Data.Generics hiding ((:*:))
import Data.Typeable
import Data.Generics.Basics
import Data.Generics.Schemes


opt :: Code -> Code
opt code 
  | code' == code = code
  | otherwise     = opt code'
  where code' = 
            everywhere (mkT removeIdLoop) $
            everywhere (mkT orderMath   ) $
            everywhere (mkT simplifyMath) $
            code

-- we're often left with things like:
--    For i = ... do
--      x_{..., i, ...} += v * [i=q]
-- (or where we have assignment instead of increment)
--
-- we want to replace these with
--    x_{..., q, ...} += v
removeIdLoop :: Statement -> Statement
removeIdLoop s@(Loop idx _ _ [Assn (V var sub) e]) =
  case (findIndices (==(Var (V idx []))) sub, extractEquality idx e) of
    ([subIdx], (Just id,Just enew)) -> Assn (V var (replaceNth sub subIdx id)) enew
    _ -> s
removeIdLoop s@(Loop idx _ _ [Incr (V var sub) e]) =
  case (findIndices (==(Var (V idx []))) sub, extractEquality idx e) of
    ([subIdx], (Just id,Just enew)) -> Incr (V var (replaceNth sub subIdx id)) enew
    _ -> s
removeIdLoop s = s

-- orderMath assembles expressions on the "proper" side to be simplified
orderMath (Fun f0 [Fun f1 [a,b], c])
  | isOp f0 && f0 == f1 = (Fun f0 [Fun f1 [a',b'], c'])
  where 
    isOp x = x `elem` [".+",".-",".*","./","+","-","*","/"]
    [a',b',c'] = sort [a,b,c]
orderMath (Fun f0 [a, Fun f1 [b,c]])
  | isOp f0 && f0 == f1 = (Fun f0 [Fun f1 [a',b'], c'])
  where 
    isOp x = x `elem` [".+",".-",".*","./","+","-","*","/"]
    [a',b',c'] = sort [a,b,c]
orderMath e = e

-- simplifyMath removes multiplication by 1, exponentiation by 1 and zero, and addition of zero
simplifyMath (Fun  "*" [Con (C.I 1),e]) = e
simplifyMath (Fun ".*" [Con (C.I 1),e]) = e
simplifyMath (Fun  "*" [e,Con (C.I 1)]) = e
simplifyMath (Fun ".*" [e,Con (C.I 1)]) = e
simplifyMath (Fun  "*" [Con (C.R 1),e]) = e
simplifyMath (Fun ".*" [Con (C.R 1),e]) = e
simplifyMath (Fun  "*" [e,Con (C.R 1)]) = e
simplifyMath (Fun ".*" [e,Con (C.R 1)]) = e

simplifyMath (Fun  "^" [e,Con (C.I 1)]) = e
simplifyMath (Fun ".^" [e,Con (C.I 1)]) = e
simplifyMath (Fun  "^" [e,Con (C.R 1)]) = e
simplifyMath (Fun ".^" [e,Con (C.R 1)]) = e

simplifyMath (Fun  "+" [Con (C.I 0),e]) = e
simplifyMath (Fun ".+" [Con (C.I 0),e]) = e
simplifyMath (Fun  "+" [e,Con (C.I 0)]) = e
simplifyMath (Fun ".+" [e,Con (C.I 0)]) = e
simplifyMath (Fun  "+" [Con (C.R 0),e]) = e
simplifyMath (Fun ".+" [Con (C.R 0),e]) = e
simplifyMath (Fun  "+" [e,Con (C.R 0)]) = e
simplifyMath (Fun ".+" [e,Con (C.R 0)]) = e

simplifyMath (Fun  "-" [e,Con (C.I 0)]) = e
simplifyMath (Fun ".-" [e,Con (C.I 0)]) = e
simplifyMath (Fun  "-" [e,Con (C.R 0)]) = e
simplifyMath (Fun ".-" [e,Con (C.R 0)]) = e

simplifyMath (Fun  "+" [Con c, Con d])  = Con (c + d)
simplifyMath (Fun ".+" [Con c, Con d])  = Con (c + d)
simplifyMath (Fun  "-" [Con c, Con d])  = Con (c - d)
simplifyMath (Fun ".-" [Con c, Con d])  = Con (c - d)
simplifyMath (Fun  "*" [Con c, Con d])  = Con (c * d)
simplifyMath (Fun ".*" [Con c, Con d])  = Con (c * d)
simplifyMath (Fun  "/" [Con c, Con d])  = Con (c / d)
simplifyMath (Fun "./" [Con c, Con d])  = Con (c / d)
simplifyMath (Fun  "^" [Con c, Con d])  = Con (c ** d)
simplifyMath (Fun ".^" [Con c, Con d])  = Con (c ** d)

simplifyMath e = e

-- extractEquality idx e returns (id,enew) if the only occurance of
-- idx in e is in a product with an equality.  id is the thing it is
-- in equality to, enew is the entire expression with the equality
-- replaced by "1"
extractEquality idx e@(Var (V id sl)) | id == idx = (Nothing,Nothing)
                                      | otherwise = (Nothing,Just e)
extractEquality idx e@(Con _) = (Nothing, Just e)
extractEquality idx e@(Fun "=" [Var (V id []), eq])
  | id == idx = case extractEquality idx eq of
                  (Nothing, Just eq') -> (Just eq', Just (Con (C.R 1)))
                  _ -> (Nothing, Just e)
extractEquality idx e@(Fun fn [a,b])
  | fn `elem` ["+","-","*","/",".+",".-",".*","./"] =
  case (extractEquality idx a, extractEquality idx b) of
    ((Nothing,Just a'), (Nothing,Just b')) -> (Nothing, Just (Fun fn [a',b']))
    ((Just id,Just a'), (Nothing,Just b')) -> (Just id, Just (Fun fn [a',b']))
    ((Nothing,Just a'), (Just id,Just b')) -> (Just id, Just (Fun fn [a',b']))

    ((Nothing,Just _ ), (Nothing,Nothing)) -> (Nothing, Just e)
    ((Nothing,Nothing), (Nothing,Just _ )) -> (Nothing, Just e)

    ((Just id,Just a'), (Nothing,Nothing)) -> (Just id, Just (Fun fn [a',b ]))
    ((Nothing,Nothing), (Just id,Just b')) -> (Just id, Just (Fun fn [a ,b']))

    _ -> (Nothing,Nothing)
extractEquality idx _ = (Nothing,Nothing)


